#ifndef ZAD2_GRAPH_H
#define ZAD2_GRAPH_H
#include <unordered_set>
template<typename T>
class Graph {
public:
	Graph() { };
	void add(T &i, T &j) {
		vertices[i];
		vertices[j];
		edgesSet.emplace(i, j);
		edgesSet.emplace(j, i);
	}

	set<pair<T, T>> edgesSet;
	map<T, int> vertices;
	vector<T> reverseVertices;
	vector<unordered_set<int>> edges;
	vector<int> order;
	vector<int> color;
	bool chordal = false;

	auto getEdges() const {
		return edgesSet;
	}

	vector<T> getOrder() const {
		vector<T> ret;
		for (auto i : order)
			ret.emplace_back(reverseVertices[i]);
		return ret;
	}

	void buildGraph() {
		int n = 0;
		reverseVertices.reserve(vertices.size());
		for (auto &i : vertices) {
			i.second = n++;
			reverseVertices.emplace_back(i.first);
		}

		edges.resize(vertices.size());
		for (auto e : edgesSet) {
			auto a = vertices[e.first], b = vertices[e.second];
			edges[a].emplace(b);
		}
	}

	void initOrder() {
		if (vertices.empty()) return;
		vector<int> card;
		vector<unordered_set<int>> lambda;
		int n = vertices.size();
		card.resize(n);
		lambda.resize(n);
		for (int i = 0; i < n; i++) {
			lambda[0].emplace(i);
		}

		while (!lambda.empty()) {
			if (lambda.back().empty()) {
				lambda.pop_back();
				continue;
			}
			auto v = *lambda.back().begin();
			order.push_back(v);
			for (auto x : edges[v]) {
				auto &c = card[x];
				if (c >= 0) {
					lambda[c].erase(x);
					c++;
					if (lambda.size() <= c)
						lambda.emplace_back();
					lambda[c].emplace(x);
				}
			}
			lambda[card[v]].erase(v);
			card[v] = -1;
		}
	}

	void initChordal() {
		unordered_set<int> visited;
		for (auto i : order) {
			unordered_set<int> neighbors;
			for (auto x : edges[i]) {
				if (visited.count(x)) {
					neighbors.emplace(x);
				}
			}
			for (auto x : neighbors) {
				for (auto y : neighbors) {
					if (x == y) continue;
					if (!edges[x].count(y) || !edges[y].count(x)) {
						chordal = false;
						return;
					}
				}
			}
			visited.emplace(i);
		}
		chordal = true;
	}

	vector<int> spilledVertices;
	int highestColor;
	map<int, int> colorUses;
	vector<int> colorCount;

	void initColoring(int maxC) {
		highestColor = -1;
		color.resize(vertices.size());
		for (auto &i : color) i = -1;
		for (auto c : order) {
			bool used[maxC];
			for (auto x : edges[c]) {
				if (color[x] == -1) continue;
				used[color[x]] = true;
			}
			for (int col = 0; col < maxC; col++) {
				if (!used[col]) {
					color[c] = col;
					break;
				}
			}
			if (color[c] == -1) {
				disconnectNode(c);
				spilledVertices.emplace_back(c);
			} else {
				highestColor = max(highestColor, color[c]);
			}
		}
		chordal = true;
	}

	void disconnectNode(int node) {
		for (auto x : edges[node]) {
			edges[x].erase(node);
		}
	}

	void setup() {
		if (vertices.empty()) {
			throw std::runtime_error("ayy lmao empty vertices");
		}
		buildGraph();
		initOrder();
		initChordal();
	}
};

#endif //ZAD2_GRAPH_H
