#include "RegisterAllocator.h"
#include "BasicBlock.h"
#include "setOverloads.h"
#include <fstream>

void RegisterAllocator::analyseLive() {
	for (const auto& v : vars) {
		if (!v->info) continue;
		for (const auto& q : v->writes) {
			q->block->flow.addDef(v);
		}
		for (const auto& q : v->uses) {
			q->block->flow.addUse(v);
		}
	}

	for (const auto& b : blocks) {
		for (const auto& phiMap : b->phi->phi) {
			for (const auto& phiRow : phiMap) {
				b->flow.def.emplace(phiRow.first);
			}
		}
	}

	while (true) {
		bool changed = false;
		for (const auto& b : blocks) {
			set<VariablePtr> pIn, pOut, diff;
			auto &f = b->flow;
			pIn.swap(f.in);
			pOut.swap(f.out);
			f.in = f.use + (pOut - f.def);
			for (const auto& succ : b->out) {
				f.out += succ->flow.inForBlock(b);
			}
			if (f.in != pIn || f.out != pOut) changed = true;
		}
		if (!changed) break;
	}

	if (!blocks.empty() && !blocks[0]->flow.in.empty())
		throw ParseError("Variable definition not found - first block expects input");
}

void RegisterAllocator::allocate() {
	analyseLive();
	buildGraph();
}

void printGraph(const Graph<VariablePtr> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first->name << " " << v.second->name << endl;
	}
	f.close();
}

void printGraph(const Graph<std::string> &graph, const std::string& filename) {
	ofstream f(filename);
	for (const auto &v : graph.getEdges()) {
		f << v.first << " " << v.second << endl;
	}
	f.close();
}

void RegisterAllocator::buildGraph() {
	bool anyAlive = false;
	for (const auto& b : blocks) {
		set<VariablePtr> alive = b->flow.in;
		for (const auto& q : b->quads) {
			alive += to_set(q->definitions());
			vector<VariablePtr> dead;
			for (const auto& var : alive) {
				// this quad is the last quad that uses var
				if ((var->uses.empty() || var->uses.back() == q) && !b->flow.out.contains(var)) {
					dead.emplace_back(var);
				}
			}
			alive -= to_set(dead);
			q->aliveAfter = alive;
			anyAlive |= alive.size() >= 2;
			for (auto i : alive) {
				for (auto j : alive) {
					if (i != j) {
						graph.add(i, j);
						g2.add(i->name, j->name);
					}
				}
			}
		}
	}
	if (!anyAlive) return;
	graph.setup();
	g2.setup();
	assert(graph.chordal == g2.chordal);
	if (!g2.chordal) {
		printGraph(graph, "./mygraph");
		printGraph(g2, "./myg2");
		throw ParseError("graph not chordal!");
	}
}


